'''
..
  $Id$
  
  GTK python tools of the clazzes.org project
  http://www.clazzes.org
 
  Created: 21.09.2018
 
  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at
 
      http://www.apache.org/licenses/LICENSE-2.0
 
  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License. 
'''

from clazzes.gtk import render

import gi
import bisect

gi.require_version('Gdk', '3.0')
gi.require_version('Gtk', '3.0')
from gi.repository import Gdk
from gi.repository import Gtk
from gi.repository import cairo
from gi.repository import PangoCairo
from gi.repository import GObject

import math
import logging

log = logging.getLogger(__name__)

class MultiSelectionView(Gtk.DrawingArea):
    '''
       A view of multi-selected labels, which may be deselected by the user.
       
       This widget serves a the display widget for a multi-selection combo box.
    '''
    def __init__(self,model:Gtk.TreeModel,text_column=0,padding=5,border=5,corner_radius=7,label_color=Gdk.color_parse("#00ff00"),editable=True):
        '''
           Constructor for the multi-selection view.
        
           :param model: The tree model to use.
           :param text_column: The column of the model to show inside the labels. 
           :param padding: The amount of padding between the border and the label.
           :param padding: The amount of padding inside the label.
           :param corner_radius: The corner radius.
           :param label_color: The color of the shown labels.          
           :param editable: If True, user interaction is allowed.
        '''
        Gtk.DrawingArea.__init__(self)
        
        ctx = self.get_style_context()
        ctx.add_class(Gtk.STYLE_CLASS_BACKGROUND)
        ctx.add_class(Gtk.STYLE_CLASS_VIEW)
        ctx.add_class(Gtk.STYLE_CLASS_FRAME)

        self._selection = []
        
        self.model = model
        self.border = border
        self.padding = padding
        self.corner_radius = corner_radius
        self.label_color = label_color
        
        self.text_column = text_column
        
        self._editable = editable

        if not self._editable:        
            ctx.add_class(Gtk.STYLE_CLASS_READ_ONLY)
        
        # stored 5-tuple positions of labels as returned by render.draw_rounded_label()
        self._label_geometries = None
        # close icon of which label has bee hovered
        self._hovered_idx = None
        
        layout = self.create_pango_layout("M")
        
        extents = layout.get_pixel_extents()
        
        self.set_size_request(extents[1].width*20,extents[1].height+2*self.padding+2*self.border)

        self.add_events(Gdk.EventMask.POINTER_MOTION_MASK|Gdk.EventMask.BUTTON_PRESS_MASK|Gdk.EventMask.BUTTON_RELEASE_MASK)

        self.connect("motion-notify-event", self._on_pointer_motion)
        self.connect("leave-notify-event",self._on_leave)
        self.connect("button-release-event",self._on_button_release)

    @GObject.Signal(name="selection-changed")
    def on_selection_changed(self):
        '''
           The "selection-changed" signal is emitted, when the selection of the
           view has changed. Call get_eslection() on the widget to get the
           actual list of select tree paths.
        '''
        if self._hovered_idx is not None:
            self._hovered_idx = None
    
        self.queue_draw()
        
    @GObject.Signal(name="path-toggled",arg_types=(Gtk.TreePath.__gtype__,bool))
    def on_path_toggled(self,path:Gtk.TreePath,state:bool):
        '''
           The "path-toggled" signal is emitted, when the selection state of path has changed.
           
           :param path: The path, which has been toggled.
           :param state: Whether the path is selected.
        '''
        if log.getEffectiveLevel() <= logging.DEBUG:
            log.debug("MultiSelectionView: Path [%s] toggled to state [%s]"%(path,state))

    def select_path(self,path:Gtk.TreePath):
        '''
           Add the given path to the selection.
           
           :param path: A Gtk.TreePath to select.
           :return: Whether this path has been newly selected. False is returned,
                    if the given path has already been selected.
        '''
        idx = bisect.bisect_left(self._selection,path)
        
        if idx < len(self._selection) and self._selection[idx] == path:
            log.warn("Path [%s] is already selected."%(path,))
            return False
        
        self._selection.insert(idx,path)
        self.emit("path-toggled",path,True)
        self.emit("selection-changed")
        return True
    
    def set_editable(self,editable:bool):
        '''
           Set whether this widget accept user input.
           
           :param editable: If True, user interaction is allowed.
        '''
        if editable != self._editable:
        
            ctx = self.get_style_context()
        
            self._editable = editable
        
            if self._editable:
                ctx.remove_class(Gtk.STYLE_CLASS_READ_ONLY)
            else:
                ctx.add_class(Gtk.STYLE_CLASS_READ_ONLY)
                
                if self._hovered_idx is not None:
                    self._hovered_idx = None
                    self.queue_draw()
            
    def get_editable(self):
        '''
           @return: Whether user interaction is currently allowed.
        '''
        return self._editable

    def toggle_path(self,path:Gtk.TreePath,state:bool = None):
        '''
           Toggle the selection state of the given path.
           
           :param path: A Gtk.TreePath to toggle.
           :param state: The state to set or None to alter the current
                         state of the path.
           :return: The final selection state of the given path.
        '''
        
        if state is None:      
            idx = bisect.bisect_left(self._selection,path)
        
            if idx < len(self._selection) and self._selection[idx] == path:
                self._selection.pop(idx)
                ret = False
            else:
                self._selection.insert(idx,path)
                ret = True
            
            self.emit("path-toggled",path,ret)
            self.emit("selection-changed")
            return ret
        elif state:
            self.select_path(path)
            return True
        else:
            self.deselect_path(path)
            return False
           
    def deselect_path(self,path:Gtk.TreePath):
        '''
           Remove the given path to the selection.
           
           :param path: A Gtk.TreePath to select.
           :return: Whether this path has been deselected. False is returned,
                    if the given path has already been deselected.
        '''
        
        if not self.is_path_selected(path):
            log.warn("Path [%s] is not selected."%(path,))
            return False
        
        self._selection.remove(path)
        self.emit("path-toggled",path,False)
        self.emit("selection-changed")
        return True

    def clear_selection(self):
        '''
           Remove all paths from the selection.
           
           :param path: A Gtk.TreePath to select.
           :return: Whether this path has been deselected. False is returned,
                    if the given path has already been deselected.
        '''
        paths_deselected = self._selection.copy()
        self._selection.clear()
        
        for path in paths_deselected:
            self.emit("path-toggled",path,False)
        self.emit("selection-changed")
    
    def is_path_selected(self,path:Gtk.TreePath):
        '''
            :param path: The path to query the selection state of.
            :return: Whether the path is currently selected. 
        '''
        
        idx = bisect.bisect_left(self._selection,path)        
        return idx < len(self._selection) and self._selection[idx] == path

    def get_selection(self):
        '''
           :return: A list of Gtk.TreePAth objects, which are currently selected.
        '''
        return self._selection.copy()
    
    def get_model(self):
        '''
           :return: The model set for this view.
        '''
        return self.model
    
    def _on_leave(self,a,event):  # @UnusedVariable
        if self._hovered_idx is not None:
            self._hovered_idx = None
            self.queue_draw()
    
        return True
    
    def _on_button_release(self,a,event):  # @UnusedVariable
        
        if event.button == 1 and self._hovered_idx is not None:
            
            path = self._selection.pop(self._hovered_idx)
            self.emit("path-toggled",path,False)
            self.emit("selection-changed")
            
    def _on_pointer_motion(self,a,event):  # @UnusedVariable
        
        if self._editable and self._label_geometries is not None:
            for i in range(len(self._label_geometries)):
                
                geom = self._label_geometries[i]
                
                d = math.hypot(event.x-geom[4],event.y-geom[5])
        
                if d <= geom[6]*0.5 + 1:
                    
                    if self._hovered_idx != i:
                        self._hovered_idx = i
                        self.queue_draw()
                        #self.get_window().set_cursor(Gdk.Cursor.new(Gdk.CursorType.CROSS))
                        
                    return True
        
        if self._hovered_idx != None:
            self._hovered_idx = None
            self.queue_draw()
    
        #self.get_window().set_cursor(Gdk.Cursor.new(Gdk.CursorType.CENTER_PTR))
        return True
    
    def do_draw(self,cr:cairo.Context):  # @UnusedVariable
        
        w = self.get_allocated_width()
        h = self.get_allocated_height()
            
        cr.save()

        try:
            
            Gtk.render_background(self.get_style_context(),cr,0,0,w,h)
            Gtk.render_frame(self.get_style_context(),cr,0,0,w,h)
            
            min_w = 2*h
            max_n = (w - self.padding) // (min_w + self.padding)
            
            if len(self._selection) > max_n:
                remainders = self._selection[max_n:]
                shown_paths = self._selection[:max_n]
            else:
                remainders = None
                shown_paths = self._selection
            
            xpos = self.border
            
            self._label_geometries = []
            
            if remainders is not None:
                ellipsis = self.create_pango_layout("\u2026")
                extents = ellipsis.get_pixel_extents()[1]
                ell_w = extents.width + self.padding
                ell_h = extents.height
            else:
                ellipsis = None
                ell_w = 0
                ell_h = 0
            
            for path in shown_paths:
            
                it = self.model.get_iter(path)
                msg, = self.model.get(it,self.text_column)

                n = len(self._label_geometries)

                ww = (w - xpos - self.border - ell_w) / (len(shown_paths)-n) ** 0.85

                geometry = render.draw_rounded_label(self,cr,self.label_color,msg,xpos,self.border,ww,h-2*self.border,focused=self._hovered_idx == n)
            
                self._label_geometries.append(geometry)
                xpos += geometry[2] + self.padding
            
            if ellipsis is not None:
                cr.move_to(xpos,(h-ell_h)/2)
                cr.set_source_rgb(0.0, 0.0, 0.0)
    
                PangoCairo.show_layout(cr,ellipsis)
            
        finally:
            cr.restore()
            
        return True

def get_menu_from_treemodel(treemodel,column=0,function=None,editable = True):
    menu = Gtk.Menu()
    treeiter = treemodel.get_iter_first()
    while treeiter != None:
        item = Gtk.CheckMenuItem(label=treemodel.get_value(treeiter,column))
        item.set_sensitive(editable)
        menu.append(item)
        if function is not None:
            item.connect("toggled",function,treemodel.get_path(treeiter))
        treeiter = treemodel.iter_next(treeiter)
    menu.set_halign(Gtk.Align.END)
    menu.show_all()
    return menu

class MultiSelectionCombo(Gtk.HBox):
    '''
       A multi-seclection combo box
    '''
    def __init__(self,model:Gtk.TreeModel,text_column=0,padding=5,border=5,corner_radius=7,label_color=Gdk.color_parse("#00ff00"),editable=True):
        '''
           Constructor for the multi-selection combo box.
        
           :param model: The tree model to use.
           :param text_column: The column of the model to show inside the labels. 
           :param padding: The amount of padding between the border and the label.
           :param padding: The amount of padding inside the label.
           :param corner_radius: The corner radius.
           :param label_color: The color of the shown labels.
           :param editable: If True, user interaction is allowed.
        '''
        
        Gtk.HBox.__init__(self)
        
        self.view = MultiSelectionView(model,text_column,padding,border,corner_radius,label_color,editable)
        
        self.get_selection = self.view.get_selection
        self.get_model = self.view.get_model
        self.is_path_selected = self.view.is_path_selected
        self.select_path = self.view.select_path
        self.deselect_path = self.view.deselect_path
        self.toggle_path = self.view.toggle_path
        self.clear_selection = self.view.clear_selection
        self.get_editable = self.view.get_editable

        self.pack_start(self.view,True,True,0)

        self.menu = get_menu_from_treemodel(model,text_column,self._on_item_toggled,editable)

        self.button = Gtk.MenuButton(popup=self.menu)
        self.pack_start(self.button,False,False,0)
        self.button.set_align_widget(self)
        
        self.view.connect("path-toggled",self._on_path_toggled)
        self.view.connect("selection-changed",self._on_selection_changed)


    def set_editable(self,editable):
        '''
           Set whether this widget accept user input.
           
           :param editable: If True, user interaction is allowed.
        '''
        if editable != self.get_editable():
        
            self.view.set_editable(editable)
            
            for item in self.menu.get_children():
                item.set_sensitive(editable)
        
        

    @GObject.Signal(name="path-toggled",arg_types=(Gtk.TreePath.__gtype__,bool))
    def on_path_toggled(self,path:Gtk.TreePath,state:bool):
        '''
           The "path-toggled" signal is emitted, when the selection state of path has changed.
           
           :param path: The path, which has been toggled.
           :param state: Whether the path is selected.
        '''
        
        if log.getEffectiveLevel() <= logging.DEBUG:
            log.debug("MultiSelectionCombo: Path [%s] toggled to state [%s]"%(path,state))

    @GObject.Signal(name="selection-changed")
    def on_selection_changed(self):
        '''
           The "selection-changed" signal is emitted, when the selection of the
           view has changed. Call get_eslection() on the widget to get the
           actual list of select tree paths.
        '''
        if log.getEffectiveLevel() <= logging.DEBUG:
            log.debug("MultiSelectionCombo: Selection changed to %s"%(self.get_selection(),))

    def _on_selection_changed(self,w):  # @UnusedVariable
        # forward selection changed events from view.
        self.emit("selection-changed")

    def _on_item_toggled(self,menuitem,path):
        
        state = menuitem.get_active()
        
        if self.view.is_path_selected(path) != state:
            self.view.toggle_path(path,state)
        
    def _on_path_toggled(self,w,path,state):  # @UnusedVariable
    
        idx = path.get_indices()[0]
        
        item = self.menu.get_children()[idx]
        if item.get_active() != state:
            item.set_active(state)
            
        self.emit("path-toggled",path,state)