'''
..
  $Id$
  
  GTK python tools of the clazzes.org project
  http://www.clazzes.org
 
  Created: 21.09.2018
 
  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at
 
      http://www.apache.org/licenses/LICENSE-2.0
 
  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License. 
'''

import gi

gi.require_version('Gtk', '3.0')
from gi.repository import Gtk
from gi.repository import GObject

import logging
from collections.abc import Sequence

log = logging.getLogger(__name__)

class ClazzesDlg(Gtk.Window):
    '''
       An implementation, that follows the API of Gtk.Dialog, but allows
       to intercept the response signal by adding an event handler, which
       might return True to indicate, that the default behavior of closing
       the dialog window and returning from run() should be skipped.
       
       Moreover, buttons added by add_button() with do_connect=False can be connected
       to a user-defined signal, which starts an asynchronous operation or the like.
       The callback from the asynchronous operation might then emit the "response"
       signal in order to terminate the dialog.
       
       Buttons associated with a response type might be enabled or disabled through
       set_response_sensitive(), which also takes lists or tuples of responses instead
       of a single response type.
    '''

    @GObject.Signal(name="response",flags=GObject.SignalFlags.RUN_LAST, return_type=bool,
                    arg_types=(int,),
                    accumulator=GObject.signal_accumulator_true_handled)
    def on_response(self,response):
        
        if not self.get_response_sensitive(response):
            log.warn("ClazzesDlg response [%s] triggered, although it is set to insensitive."%(response,))
            return True
            
        self.response = response
        Gtk.main_quit()
        return False

    def _emit_response(self,*args):        
        return self.emit("response",args[-1])

    def __init__(self,title=None,parent=None,modal=True,layout_style=Gtk.ButtonBoxStyle.END,padding=3,connect_delete_event=True):
        '''
           Constructor for a ClazzesDlg.
           :param title: The window title.
           :param parent: The optional modal parent.
           :param modal: Whether the dialog is modal.
           :param layout_style: The layout style of the embedded ButtonBox.
           :param padding: The padding used to place buttons.
           :param connect_delete_event: If True, "delete-event" is connected to emit
                         the "response" signal with parameter Gtk.ResponseType.DELETE_EVENT
        '''
        Gtk.Window.__init__(self,title=title,modal=modal)

        if parent is not None:
            self.set_transient_for(parent)

        self.padding = padding

        self.vbox = Gtk.VBox(border_width=self.padding)
        
        self.buttonbox = Gtk.ButtonBox(layout_style=layout_style,spacing=self.padding)
        
        self.vbox.pack_end(self.buttonbox,False,False,0)
        
        self.response = Gtk.ResponseType.NONE
        
        self.response_buttons = {}
        self.response_sensitivity = {}
        
        self.add(self.vbox)
        if connect_delete_event:
            self.connect("delete-event",self._emit_response,Gtk.ResponseType.DELETE_EVENT)
    
    def add_content(self,widget):
        '''
           Add the main widget, for the most part a VBOx or a Grid is used.
           :param widget: The central dialog widget.
        '''
        self.vbox.pack_start(widget,True,True,self.padding)

    def add_button(self,button,response=None,do_connect=True):
        '''
           Add a button to the dialog.
           
           :param button: A Gtk.Button to add.
           :param response: The Gtk.ResponseType of the button. This is used for
                         set_response_type_sensitive().
           :param do_connect: If True (the default), the button is connected to
                         emit the "response" signal with the given response type.
        '''
        self.buttonbox.add(button)
        if response is not None:
            
            if do_connect:
                button.connect("clicked",self._emit_response,response)
            
            resp_buttons = self.response_buttons.get(response)
            if resp_buttons is None:
                resp_buttons = [button]
                self.response_buttons[response] = resp_buttons
            else:
                resp_buttons.append(button)            
    
    def set_response_sensitive(self,response,sensitive):
        '''
           Enable/disable a given response type or a sets of response types.
           
           :param response: A Gtk.ReponseType or a tuple or list of Gtk.ResponseType instances.
           :param sensitive: Whether to enable or disable the response type(s).
        '''
        
        if isinstance(response,Sequence):
            for r in response:
                self.set_response_sensitive(r,sensitive)
        else:
            self.response_sensitivity[int(response)] = sensitive
            resp_buttons = self.response_buttons.get(response)
            
            if resp_buttons is not None:
                for button in resp_buttons:
                    button.set_sensitive(sensitive)
                
    def get_response_sensitive(self,response):
        '''
           :return: Whether the given response type is currently enabled or disabled. 
        '''
        return self.response_sensitivity.get(int(response),True)
        
    def run(self):
        '''
           Run the dialog and return, when the "response" signal has been processed
           with a return value of False, i.e. no handler intercepted the standard
           behavior.
            
           :return: The final response type, which might be Gtk.ResponseType.DELETE_EVENT
        '''
        self.show_all()
        Gtk.main()
        return self.response
        