'''
..
  $Id$
  
  GTK python tools of the clazzes.org project
  http://www.clazzes.org
 
  Created: 21.09.2018
 
  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at
 
      http://www.apache.org/licenses/LICENSE-2.0
 
  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License. 
'''

import gi

gi.require_version('Gdk', '3.0')
gi.require_version('Gtk', '3.0')
gi.require_version('PangoCairo', '1.0')
from gi.repository import Gdk
from gi.repository import Gtk
from gi.repository import cairo
from gi.repository import PangoCairo
from gi.repository import Pango

import math

ALPHA = (math.sqrt(2)-1) * 4/3

def ell_arc_to(cr: cairo.Context, dx1, dy1, dx2, dy2):
    '''
       Draw an elliptical arc from the current point of the cairo context
       with two relative half-axes.
       
       After this operation, that current point is moved by dx1+dx2 and
       dy1+dy2, respectively.
       
       :param cr: A cairo context to draw to.
       :param dx1: The x advance of the first half-axis.
       :param dy1: The y advance of the first half-axis.
       :param dx2: The x advance of the second half-axis.
       :param dy2: The y advance of the second half-axis.
    '''
    cr.rel_curve_to(dx1*ALPHA,dy1*ALPHA,dx1+dx2*(1-ALPHA),dy1+dy2*(1-ALPHA),dx1+dx2,dy1+dy2)
    

def rounded_rectangle(cr:cairo.Context,x,y,w,h,r):
    '''
       Draw a rounded rectangle to the given cairo context.
       
       If the corner radius is smaller than h/2 or w/2, the corner radius is reduced
       to the smaller value of these.
       
       :param cr: A cairo context to draw to.
       :param x: The x coordinate of the upper left corner of the rectangle.
       :param y: The y coordinate of the upper left corner of the rectangle.
       :param w: The width of the rectangle.
       :param h: The height of the rectangle.
       :param r: The corner radius.
    '''
    
    rr = min(r,w/2,h/2)
    
    cr.move_to(x+w-rr,y)
    ell_arc_to(cr,rr,0,0,rr)
    cr.line_to(x+w,y+h-rr)
    ell_arc_to(cr,0,rr,-rr,0)
    cr.line_to(x+rr,y+h)
    ell_arc_to(cr,-rr,0,0,-rr)
    cr.line_to(x,y+rr)
    ell_arc_to(cr,0,-rr,rr,0)
    cr.close_path()

def draw_rounded_label(widget : Gtk.Widget, cr:cairo.Context,color:Gdk.Color,label:str,x,y,w,h,corner_radius = 7, border_width=2, padding = 5, icon_scale = 0.8, focused = False, shrink = True):
    '''
       Draw a rounded label
       
       :param widget The widget to draw on for determining the text style.
       :param cr: The cairo context which actually draws to the widget.
       :param color: The color of the label.
       :param label: The text label to render.
       :param x: The x coordinate of the upper left corner of the rectangle.
       :param y: The y coordinate of the upper left corner of the rectangle.
       :param w: The width of the rectangle. This is a maximal width, if shrink is set to True
       :param h: The height of the rectangle.
       :param corner_radius: The corner radius.
       :param border_width: The stroke width of the border line.
       :param padding: The amount of padding inside the label.
       :param icon_scale: The size of the close icon or None, if no close icon is needed. 
       :param focused: Whether to draw the close icon in focused state.
       :param shrink: WHether to shrink the label to size actually needed.
       
       :return: A 4-tuple of the actual x,y,width and height of the label drawn, if icon_scale was None.
                A 7-tuple with the actual x,y,width and height of the label and the x and y position
                of the close icon and the diameter of the close icon, if a close icon has been rendered.
    '''
    
    if icon_scale:
        icon_sz = (h - 2*padding) * icon_scale
    else:
        icon_sz = 0.0
    
    layout = widget.create_pango_layout(label)
    layout.set_width(Pango.units_from_double(w-2*padding-icon_sz))
    layout.set_height(Pango.units_from_double(h-2*padding))
    layout.set_single_paragraph_mode(True)
    layout.set_ellipsize(Pango.EllipsizeMode.MIDDLE)

    extents = layout.get_pixel_extents()
    
    width = extents[1].width
    height = extents[1].height

    if shrink:
        min_w = width + 2*padding
        if icon_sz > 0:
            min_w += padding + icon_sz
    
        ww = min(w,min_w)
    else:
        ww = w
    
    cr.set_line_width(border_width)

    rounded_rectangle(cr,x,y,ww,h,corner_radius)
    
    cr.set_source_rgba(color.red_float,color.green_float,color.blue_float,0.2)
    cr.fill_preserve()

    cr.set_source_rgb(color.red_float,color.green_float,color.blue_float)

    cr.stroke()
    
    if icon_sz > 0:
        cx = x+(ww-icon_sz-padding-width)/2
    else:
        cx = x+(ww-width)/2
    
    cr.move_to(cx,y+(h-height)/2)
    
    cr.set_source_rgb(0.0, 0.0, 0.0)
    
    PangoCairo.show_layout(cr,layout)
    
    if icon_sz > 0:
    
        cx = x+ww-padding-icon_sz/2
        cy = y+h/2
        
        if focused:
            cr.set_source_rgba(0.0,0.0,0.0,0.1)
            cr.arc(cx,cy,icon_sz/2,0,math.pi*2)
            cr.fill()
        
        cr.set_source_rgb(0.1, 0.1, 0.1)
    
        d = math.sqrt(0.125) * icon_sz - 1
    
        cr.move_to(cx-d,cy-d)
        cr.line_to(cx+d,cy+d)
        cr.stroke()
        
        cr.move_to(cx+d,cy-d)
        cr.line_to(cx-d,cy+d)
        cr.stroke()
        
        return (x,y,ww,h,cx,cy,icon_sz)
    else:
        return (x,y,ww,h)