'''
..
  $Id$
  
  GTK python tools of the clazzes.org project
  http://www.clazzes.org
 
  Created: 21.09.2018
 
  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at
 
      http://www.apache.org/licenses/LICENSE-2.0
 
  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License. 
'''

import collections
import threading
from gi.repository import GLib

import logging

log = logging.getLogger(__name__)

class Invocation:
    '''
       An invocation of a function, which might be scheduled asynchronously.

       :ivar func: The function or callable to be invoked.
       :ivar args: The list of positional arguments.
       :ivar kwargs: The dict of named arguments.
    '''
    
    __slots__ = ("func","args","kwargs")
    
    def __init__(self,func,args=None,kwargs = None):
        self.func = func
        self.args = args
        self.kwargs = kwargs
        
    def __call__(self):
        
        if self.kwargs is None:
            if self.args is None:
                return self.func()
            else:
                return self.func(*self.args)
        else:
            if self.args is None:
                return self.func(**self.kwargs)
            else:
                return self.func(*self.args,**self.kwargs)

    def __str__(self):
        return self.func.__name__

class Worker:
    '''
       A callable, which may be started in one or more threads, which holds
       a queue of to be executed invocations.
    '''

    __slots__ = ("cond","label","stopped","queue","_instance_count")
    
    def __init__(self, label):
        self.cond = threading.Condition()
        self.label = label
        self.stopped = False
        self.queue = collections.deque()
        self._instance_count = 0
    
    def stop(self):
        
        log.info("Requesting worker [%s] to stop."%self.label)
            
        with self.cond:
            self.stopped = True
            self.cond.notify_all()
    
    def callMethod(self,func,args=None,kwargs = None,onsuccess=None,onexception=None):
        '''
           Invoke a method in the background and call onsuccess or onexception callbacks
           in the GUI thread upon completion.
           
           :param func: The function or callable to be invoked.
           :param args: The list of positional arguments.
           :param kwargs: The dict of named arguments.
           :param onsuccess: A success callback, which takes the worker, the invocation and the result
                     of the invocation as parameters.
           :param onexception: An exception callback, which takes the worker, the invocation, the exception
                     of the invocation and the original onsuccess handler as parameters.
        '''
        
        with self.cond:
            self.queue.append((Invocation(func,args,kwargs),onsuccess,onexception))
            self.cond.notify()
    
    def invoke(self,invocation : Invocation,onsuccess=None,onexception=None):
        '''
           Invoke a method in the background and call onsuccess or onexception callbacks
           in the GUI thread upon completion.
           
           :param invocation: The invocation to perform.
           :param onsuccess: A success callback, which takes the worker, the invocation and the result
                     of the invocation as parameters.
           :param onexception: An exception callback, which takes the worker, the invocation and the exception
                     of the invocation as parameters.
        '''
        with self.cond:
            self.queue.append((invocation,onsuccess,onexception))
            self.cond.notify()

    def __call__(self):

        with self.cond:
            self._instance_count += 1
            instance_id = self._instance_count
        
        log.info("Starting worker thread [%s-%d]."%(self.label,instance_id))

        while True:
            
            with self.cond:
                
                if self.stopped:
                    break
                
                if len(self.queue) > 0:
                
                    job = self.queue.popleft()
                
                else:
                
                    self.cond.wait()
                    
                    if self.stopped:
                        break
                    
                    if len(self.queue) > 0:
                
                        job = self.queue.popleft()
                    else:
                        job = None
 
            # execute job without holding a lock.                   
            if job is None:
                continue
            
            invocation = job[0]

            try:

                log.info("Executing job [%s]..."%(invocation,))
        
                ret = invocation()
                
                log.info("Job [%s] returned [%s]."%(invocation,type(ret)))

                if log.getEffectiveLevel() <= logging.DEBUG:
                    log.debug("Job [%s] returned value [%s]."%(invocation,ret))
                
                if job[1] is not None:
                    GLib.idle_add(job[1],self,invocation,ret)
                
            except BaseException as e:
                
                log.exception("Job [%s] finished with exception"%(invocation,))
                if job[2] is not None:
                    GLib.idle_add(job[2],self,invocation,e,job[1])
                
        log.info("Leaving worker thread [%s-%d]."%(self.label,instance_id))

THE_WORKER =  Worker("clazzes-worker")
THE_WORKER_THREADS = None

'''
   The singleton worker to be started by startTheWorker().
'''

def startTheWorker(nthreads = 1):
    
    global THE_WORKER_THREADS
    global THE_WORKER
    
    if THE_WORKER_THREADS is not None:
        raise ValueError("Worker threads are already started.")
    
    if THE_WORKER is None:
        THE_WORKER = Worker("clazzes-worker")
        
    THE_WORKER_THREADS = []
    
    for i in range(nthreads):  # @UnusedVariable
        thread = threading.Thread(target=THE_WORKER)
        thread.start()
        THE_WORKER_THREADS.append(thread)
        
def stopTheWorker():
    
    global THE_WORKER_THREADS
    global THE_WORKER

    if THE_WORKER is None:
        log.warn("stopTheWorker() called without a running worker singleton.")
        return
    
    THE_WORKER.stop()
    for thread in THE_WORKER_THREADS:
        thread.join()

    THE_WORKER_THREADS = None
